@@ -130,7 +130,7 @@ void ssl_init_Module(server_rec *s, pool *p)
     /*
      * Let us cleanup on restarts and exists
      */
-    ap_register_cleanup(p, s, ssl_init_ModuleKill, ap_null_cleanup);
+    ap_register_cleanup(p, s, ssl_init_ModuleKill, ssl_init_ModuleKill);
 
     /*
      * Any init round fixes the global config
@@ -158,7 +158,7 @@ void ssl_init_Module(server_rec *s, pool *p)
             sc->nPassPhraseDialogType = SSL_PPTYPE_BUILTIN;
 
         /* Open the dedicated SSL logfile */
-        ssl_log_open(s2, p);
+        ssl_log_open(s, s2, p);
     }
 
     /*
@@ -763,18 +763,17 @@ void ssl_init_Child(server_rec *s, pool *p)
 void ssl_init_ModuleKill(void *data)
 {
     SSLSrvConfigRec *sc;
-    server_rec *s;
+    server_rec *s = (server_rec *)data;
 
     /*
      * Drop the session cache
      */
-    ssl_scache_kill();
+    ssl_scache_kill(s);
 
     /*
      * Free the non-pool allocated structures
      * in the per-server configurations
      */
-    s = (server_rec *)data;
     for (; s != NULL; s = s->next) {
         sc = mySrvConfig(s);
         if (sc->pPublicCert[SSL_AIDX_RSA] != NULL) {
