@@ -385,7 +385,7 @@ void ssl_pphrase_Handle(server_rec *s, pool *p)
     return;
 }
 
-int ssl_pphrase_Handle_CB(char *buf, int bufsize, int w)
+int ssl_pphrase_Handle_CB(char *buf, int bufsize, int ask_twice)
 {
     SSLModConfigRec *mc = myModConfig();
     server_rec *s;
@@ -469,7 +469,7 @@ int ssl_pphrase_Handle_CB(char *buf, int bufsize, int w)
          */
         prompt = "Enter pass phrase:";
         for (;;) {
-            if ((i = EVP_read_pw_string(buf, bufsize, prompt, w)) != 0) {
+            if ((i = EVP_read_pw_string(buf, bufsize, prompt, ask_twice)) != 0) {
                 PEMerr(PEM_F_DEF_CALLBACK,PEM_R_PROBLEMS_GETTING_PASSWORD);
                 memset(buf, 0, (unsigned int)bufsize);
                 return (-1);
