@@ -96,14 +96,14 @@ void ssl_scache_init(server_rec *s, pool *p)
     return;
 }
 
-void ssl_scache_kill(void)
+void ssl_scache_kill(server_rec *s)
 {
     SSLModConfigRec *mc = myModConfig();
 
     if (mc->nSessionCacheMode == SSL_SCMODE_DBM)
-        ssl_scache_dbm_kill();
+        ssl_scache_dbm_kill(s);
     else if (mc->nSessionCacheMode == SSL_SCMODE_SHM)
-        ssl_scache_shm_kill();
+        ssl_scache_shm_kill(s);
 
 #ifdef SSL_VENDOR
     ap_hook_use("ap::mod_ssl::vendor::scache_kill",
@@ -112,12 +112,13 @@ void ssl_scache_kill(void)
     return;
 }
 
-void ssl_scache_store(server_rec *s, SSL_SESSION *pSession, int timeout)
+BOOL ssl_scache_store(server_rec *s, SSL_SESSION *pSession, int timeout)
 {
     SSLModConfigRec *mc = myModConfig();
     ssl_scinfo_t SCI;
     UCHAR buf[MAX_SESSION_DER];
     UCHAR *b;
+    BOOL rc = FALSE;
 
     /* add the key */
     SCI.ucaKey = pSession->session_id;
@@ -130,9 +131,9 @@ void ssl_scache_store(server_rec *s, SSL_SESSION *pSession, int timeout)
 
     /* and store it... */
     if (mc->nSessionCacheMode == SSL_SCMODE_DBM)
-        ssl_scache_dbm_store(s, &SCI);
+        rc = ssl_scache_dbm_store(s, &SCI);
     else if (mc->nSessionCacheMode == SSL_SCMODE_SHM)
-        ssl_scache_shm_store(s, &SCI);
+        rc = ssl_scache_shm_store(s, &SCI);
 
 #ifdef SSL_VENDOR
     ap_hook_use("ap::mod_ssl::vendor::scache_store",
@@ -142,7 +143,7 @@ void ssl_scache_store(server_rec *s, SSL_SESSION *pSession, int timeout)
     /* allow the regular expiring to occur */
     ssl_scache_expire(s, time(NULL));
 
-    return;
+    return rc;
 }
 
 SSL_SESSION *ssl_scache_retrieve(server_rec *s, UCHAR *id, int idlen)
@@ -315,7 +316,7 @@ void ssl_scache_dbm_init(server_rec *s, pool *p)
     }
     ssl_dbm_close(dbm);
 
-#ifndef WIN32
+#if !defined(OS2) && !defined(WIN32)
     /*
      * We have to make sure the Apache child processes
      * have access to the DBM file. But because there
@@ -345,19 +346,41 @@ void ssl_scache_dbm_init(server_rec *s, pool *p)
     return;
 }
 
-void ssl_scache_dbm_kill(void)
+void ssl_scache_dbm_kill(server_rec *s)
 {
-    /* nothing to be done */
+    SSLModConfigRec *mc = myModConfig();
+    pool *p;
+
+    if ((p = ap_make_sub_pool(NULL)) != NULL) {
+        /* the correct way */
+        unlink(ap_pstrcat(p, mc->szSessionCacheDataFile, SSL_DBM_FILE_SUFFIX_DIR, NULL));
+        unlink(ap_pstrcat(p, mc->szSessionCacheDataFile, SSL_DBM_FILE_SUFFIX_PAG, NULL));
+        /* the additional ways to be sure */
+        unlink(ap_pstrcat(p, mc->szSessionCacheDataFile, ".dir", NULL));
+        unlink(ap_pstrcat(p, mc->szSessionCacheDataFile, ".pag", NULL));
+        unlink(ap_pstrcat(p, mc->szSessionCacheDataFile, ".db", NULL));
+        unlink(mc->szSessionCacheDataFile);
+        ap_destroy_pool(p);
+    }
     return;
 }
 
-void ssl_scache_dbm_store(server_rec *s, ssl_scinfo_t *SCI)
+BOOL ssl_scache_dbm_store(server_rec *s, ssl_scinfo_t *SCI)
 {
     SSLModConfigRec *mc = myModConfig();
     DBM *dbm;
     datum dbmkey;
     datum dbmval;
 
+    /* be careful: do not try to store too much bytes in a DBM file! */
+#ifdef SSL_USE_SDBM
+    if ((SCI->nKey + SCI->nData) >= PAIRMAX)
+        return FALSE;
+#else
+    if ((SCI->nKey + SCI->nData) >= 1024)
+        return FALSE;
+#endif
+
     /* create DBM key */
     dbmkey.dptr  = (char *)(SCI->ucaKey);
     dbmkey.dsize = SCI->nKey;
@@ -366,7 +389,7 @@ void ssl_scache_dbm_store(server_rec *s, ssl_scinfo_t *SCI)
     dbmval.dsize = sizeof(time_t)+SCI->nData;
     dbmval.dptr  = (char *)malloc(dbmval.dsize);
     if (dbmval.dptr == NULL)
-        return;
+        return FALSE;
     memcpy((char *)dbmval.dptr, &SCI->tExpiresAt, sizeof(time_t));
     memcpy((char *)dbmval.dptr+sizeof(time_t), SCI->ucaData, SCI->nData);
 
@@ -379,7 +402,7 @@ void ssl_scache_dbm_store(server_rec *s, ssl_scinfo_t *SCI)
                 mc->szSessionCacheDataFile);
         free(dbmval.dptr);
         ssl_mutex_off(s);
-        return;
+        return FALSE;
     }
     ssl_dbm_store(dbm, dbmkey, dbmval, DBM_INSERT);
     ssl_dbm_close(dbm);
@@ -388,7 +411,7 @@ void ssl_scache_dbm_store(server_rec *s, ssl_scinfo_t *SCI)
     /* free temporary buffers */
     free(dbmval.dptr);
 
-    return;
+    return TRUE;
 }
 
 void ssl_scache_dbm_retrieve(server_rec *s, ssl_scinfo_t *SCI)
@@ -470,25 +493,52 @@ void ssl_scache_dbm_expire(server_rec *s, time_t tNow)
     SSLModConfigRec *mc = myModConfig();
     DBM *dbm;
     datum dbmkey;
-    datum dbmkeythis;
     datum dbmval;
+    pool *p;
     time_t tExpiresAt;
     int nElements = 0;
     int nDeleted = 0;
     int bDelete;
+    datum *keylist;
+    int keyidx;
+    int i;
+
+    /*
+     * Here we have to be very carefully: Not all DBM libraries are
+     * smart enough to allow one to iterate over the elements and at the
+     * same time delete expired ones. Some of them get totally crazy
+     * while others have no problems. So we have to do it the slower but
+     * more safe way: we first iterate over all elements and remember
+     * those which have to be expired. Then in a second pass we delete
+     * all those expired elements. Additionally we reopen the DBM file
+     * to be really safe in state.
+     */
+
+#define KEYMAX 1024
 
     ssl_mutex_on(s);
-    if ((dbm = ssl_dbm_open(mc->szSessionCacheDataFile,
-                            O_RDWR, SSL_DBM_FILE_MODE)) == NULL) {
-        ssl_log(s, SSL_LOG_ERROR|SSL_ADD_ERRNO,
-                "Cannot open SSLSessionCache DBM file `%s' for expiring",
-                mc->szSessionCacheDataFile);
-        ssl_mutex_off(s);
-        return;
-    }
-    dbmkey = ssl_dbm_firstkey(dbm);
-    if (dbmkey.dptr != NULL) {
-        do {
+    for (;;) {
+        /* allocate the key array in a memory sub pool */
+        if ((p = ap_make_sub_pool(NULL)) == NULL)
+            break;
+        if ((keylist = ap_palloc(p, sizeof(dbmkey)*KEYMAX)) == NULL) {
+            ap_destroy_pool(p);
+            break;
+        }
+
+        /* pass 1: scan DBM database */
+        keyidx = 0;
+        if ((dbm = ssl_dbm_open(mc->szSessionCacheDataFile,
+                                O_RDWR, SSL_DBM_FILE_MODE)) == NULL) {
+            ssl_log(s, SSL_LOG_ERROR|SSL_ADD_ERRNO,
+                    "Cannot open SSLSessionCache DBM file `%s' for scanning",
+                    mc->szSessionCacheDataFile);
+            ap_destroy_pool(p);
+            break;
+        }
+        dbmkey = ssl_dbm_firstkey(dbm);
+        while (dbmkey.dptr != NULL) {
+            nElements++;
             bDelete = FALSE;
             dbmval = ssl_dbm_fetch(dbm, dbmkey);
             if (dbmval.dsize < sizeof(time_t) || dbmval.dptr == NULL)
@@ -498,18 +548,44 @@ void ssl_scache_dbm_expire(server_rec *s, time_t tNow)
                 if (tExpiresAt >= tNow)
                     bDelete = TRUE;
             }
-            dbmkeythis = dbmkey;
-            dbmkey = ssl_dbm_nextkey(dbm);
             if (bDelete) {
-                ssl_dbm_delete(dbm, dbmkeythis);
-                nDeleted++;
+                if ((keylist[keyidx].dptr = ap_palloc(p, dbmkey.dsize)) != NULL) {
+                    memcpy(keylist[keyidx].dptr, dbmkey.dptr, dbmkey.dsize);
+                    keylist[keyidx].dsize = dbmkey.dsize;
+                    keyidx++;
+                    if (keyidx == KEYMAX)
+                        break;
+                }
             }
-        } while (dbmkey.dptr != NULL);
+            dbmkey = ssl_dbm_nextkey(dbm);
+        }
+        ssl_dbm_close(dbm);
+
+        /* pass 2: delete expired elements */
+        if ((dbm = ssl_dbm_open(mc->szSessionCacheDataFile,
+                                O_RDWR, SSL_DBM_FILE_MODE)) == NULL) {
+            ssl_log(s, SSL_LOG_ERROR|SSL_ADD_ERRNO,
+                    "Cannot re-open SSLSessionCache DBM file `%s' for expiring",
+                    mc->szSessionCacheDataFile);
+            ap_destroy_pool(p);
+            break;
+        }
+        for (i = 0; i < keyidx; i++) {
+            ssl_dbm_delete(dbm, keylist[i]);
+            nDeleted++;
+        }
+        ssl_dbm_close(dbm);
+
+        /* destroy temporary pool */
+        ap_destroy_pool(p);
+
+        if (keyidx < KEYMAX)
+            break;
     }
-    ssl_dbm_close(dbm);
     ssl_mutex_off(s);
-    ssl_log(s, SSL_LOG_TRACE, "Expired DBM-based SSL session cache: "
-            "old: %d, new: %d, removed: %d", nElements+nDeleted, nElements, nDeleted);
+
+    ssl_log(s, SSL_LOG_TRACE, "Inter-Process Session Cache (DBM) Expiry: "
+            "old: %d, new: %d, removed: %d", nElements, nElements-nDeleted, nDeleted);
     return;
 }
 
@@ -654,7 +730,7 @@ void ssl_scache_shm_init(server_rec *s, pool *p)
     return;
 }
 
-void ssl_scache_shm_kill(void)
+void ssl_scache_shm_kill(server_rec *s)
 {
     SSLModConfigRec *mc = myModConfig();
 
@@ -662,7 +738,7 @@ void ssl_scache_shm_kill(void)
     return;
 }
 
-void ssl_scache_shm_store(server_rec *s, ssl_scinfo_t *SCI)
+BOOL ssl_scache_shm_store(server_rec *s, ssl_scinfo_t *SCI)
 {
     SSLModConfigRec *mc = myModConfig();
     void *vp;
@@ -673,12 +749,12 @@ void ssl_scache_shm_store(server_rec *s, ssl_scinfo_t *SCI)
                         NULL, sizeof(time_t)+SCI->nData,
                         NULL, &vp, 1) != TABLE_ERROR_NONE) {
         ssl_mutex_off(s);
-        return;
+        return FALSE;
     }
     memcpy(vp, &SCI->tExpiresAt, sizeof(time_t));
     memcpy((char *)vp+sizeof(time_t), SCI->ucaData, SCI->nData);
     ssl_mutex_off(s);
-    return;
+    return TRUE;
 }
 
 void ssl_scache_shm_retrieve(server_rec *s, ssl_scinfo_t *SCI)
@@ -769,8 +845,8 @@ void ssl_scache_shm_expire(server_rec *s, time_t tNow)
         } while (rc == TABLE_ERROR_NONE);
     }
     ssl_mutex_off(s);
-    ssl_log(s, SSL_LOG_TRACE, "Expired SHM-based SSL session cache: "
-            "old: %d, new: %d, removed: %d", nElements+nDeleted, nElements, nDeleted);
+    ssl_log(s, SSL_LOG_TRACE, "Inter-Process Session Cache (SHM) Expiry: "
+            "old: %d, new: %d, removed: %d", nElements, nElements-nDeleted, nDeleted);
     return;
 }
 
